"""
cooldowns

Content:

* cooldown: cooldown decorator

GitHub: https://github.com/interactions-py/enhanced/blob/main/interactions/ext/enhanced/cooldowns.py

(c) 2022 interactions-py.
"""
from datetime import datetime, timedelta
from functools import wraps
from inspect import iscoroutinefunction, signature
from typing import Awaitable, Callable, Dict, Optional, Type, Union

from interactions.client.context import _Context

from interactions import Channel, Command, CommandContext, Extension, Guild, Member, User

__all__ = ("cooldown",)

NoneType: Type[None] = type(None)
_type: object = type
Coroutine = Callable[..., Awaitable]


class cooldown:
    """
    A decorator for handling cooldowns.

    Parameters for `datetime.timedelta` are `days=0, seconds=0, microseconds=0,
    milliseconds=0, minutes=0, hours=0, weeks=0`.

    ```py
    from interactions.ext.better_interactions import cooldown

    async def cooldown_error(ctx, delta):
        ...

    @client.command(...)
    @cooldown(..., error=cooldown_error, type=..., seconds=..., ...)
    async def cooldown_command(ctx, ...):
        ...
    ```

    Parameters:

    * `*delta_args: tuple[datetime.timedelta arguments]`: The arguments to pass to `datetime.timedelta`.
    * `?error: Coroutine`: The function to call if the user is on cooldown.
    * `?type: str | User | Channel | Guild = "user"`: The type of cooldown.
    * `?count: int = 1`: The number of times the user can use the command before they are on cooldown.
    * `**delta_kwargs: dict[datetime.timedelta arguments]`: The keyword arguments to pass to `datetime.timedelta`.
    """

    def __init__(
        self,
        *delta_args,
        error: Optional[Coroutine] = None,
        type: Optional[Union[str, User, Member, Channel, Guild]] = "user",
        count: int = 1,
        **delta_kwargs,
    ):
        if not (delta_args or delta_kwargs):
            raise ValueError(
                "Cooldown amount must be provided! Valid arguments and keyword arguments are listed in "
                "https://docs.python.org/3/library/datetime.html#datetime.timedelta"
            )
        if not isinstance(error, (Callable, NoneType)):
            raise TypeError("Invalid type provided for `error`! Must be a `Coroutine`!")
        if type not in {"user", User, "member", Member, "guild", Guild, "channel", Channel}:
            raise TypeError("Invalid type provided for `type`!")

        self.delta = timedelta(*delta_args, **delta_kwargs)
        self.error = error
        self.type = type

        self.last_called: Dict[str, datetime] = {}
        self.count: int = count
        self.coro_count: Dict[str, int] = {}

    def __call__(self, coro: Coroutine) -> Coroutine:
        if isinstance(coro, Command):
            raise SyntaxError("Cooldowns must go below command decorators!")

        @wraps(coro)
        async def wrapper(ctx: Union[CommandContext, Extension], *args, **kwargs):
            coro.cooldown = self
            args: list = list(args)
            _ctx: CommandContext = ctx if isinstance(ctx, _Context) else args.pop(0)
            id = self.get_id(self.type, _ctx)
            self.coro_count[id] = self.coro_count.get(id, 0) + 1
            now = datetime.now()
            unique_last_called = self.last_called.get(id)
            on_cooldown: bool = unique_last_called and (now - unique_last_called < self.delta)

            if on_cooldown and self.coro_count[id] > self.count:
                if not self.error:
                    return await _ctx.send(
                        f"This command is on cooldown for {self.delta - (now - unique_last_called)}!"
                    )
                return (
                    (
                        await self.error(_ctx, self.delta - (now - unique_last_called))
                        if iscoroutinefunction(self.error)
                        else self.error(_ctx, self.delta - (now - unique_last_called))
                    )
                    if len(signature(self.error).parameters) == 2
                    else (
                        await self.error(ctx, _ctx, self.delta - (now - unique_last_called))
                        if iscoroutinefunction(self.error)
                        else self.error(ctx, _ctx, self.delta - (now - unique_last_called))
                    )
                )

            self.last_called[id] = now
            self.coro_count[id] = 1 if self.coro_count[id] > self.count else self.coro_count[id]
            if isinstance(ctx, _Context):
                return await coro(_ctx, *args, **kwargs)
            return await coro(ctx, _ctx, *args, **kwargs)

        return wrapper

    def reset(self, id: Optional[str] = None):
        """
        Resets the cooldown.

        ```py
        @client.command(...)
        @cooldown(..., type=..., seconds=..., ...)
        async def cooldown_command(ctx, ...):
            cooldown_command.cooldown.reset(cooldown.get_id("user", ctx))
        ```

        Parameters:

        * `?id: str`: The id of the cooldown to reset. If not provided, all cooldowns are reset.
        """
        if id:
            self.last_called.pop(id)
            self.coro_count.pop(id)
        else:
            self.last_called = {}
            self.coro_count = {}

    @staticmethod
    def get_id(
        type: Optional[Union[str, User, Member, Channel, Guild]], ctx: CommandContext
    ) -> str:
        """
        Returns the appropriate ID for the type provided.

        Parameters:

        * `?type: str | User | Member | Channel | Guild`: The type of cooldown.
        * `ctx: CommandContext`: The context to get the id from.
        """
        type = type.lower() if isinstance(type, str) else type

        if type == "user" or type is User:
            return str(ctx.user.id)
        if type == "member" or type is Member:
            return f"{ctx.guild.id}:{ctx.author.id}"
        if type == "channel" or type is Channel:
            return str(ctx.channel_id)
        if type == "guild" or type is Guild:
            return str(ctx.guild_id)
        raise TypeError("Invalid type provided for `type`!")



"""
components

Content:

* ActionRow: enhanced action row
* Button: enhanced button
* SelectOption: enhanced select option
* SelectMenu: enhanced select menu
* TextInput: enhanced text input
* Modal: enhanced modal
* spread_to_rows: spread components to rows

GitHub: https://github.com/interactions-py/enhanced/blob/main/interactions/ext/enhanced/components.py

(c) 2022 interactions-py.
"""
from typing import List, Optional, Union

from interactions import ActionRow as AR
from interactions import Button as B
from interactions import ButtonStyle, Emoji
from interactions import Modal as M
from interactions import SelectMenu as SM
from interactions import SelectOption as SO
from interactions import TextInput as TI
from interactions import TextStyleType as TST

from ._logging import get_logger

__all__ = (
    "ActionRow",
    "Button",
    "SelectOption",
    "SelectMenu",
    "TextInput",
    "Modal",
)

log = get_logger("components")


def ActionRow(*args: Union[B, SM, TI]) -> AR:
    """
    A helper function that passes arguments to `ActionRow`.

    Previous:

    ```py
    row = ActionRow(components=[...])
    ```

    Now:

    ```py
    row = ActionRow(...)
    ```

    Parameters:

    * `*args: tuple[Button | SelectMenu | TextInput]`: The components to add to the `ActionRow`.

    Returns:

    `ActionRow`
    """
    log.debug(f"Creating ActionRow with {list(args)}")
    return AR(components=list(args))


def Button(
    style: Union[ButtonStyle, int],
    label: str,
    *,
    custom_id: Optional[str] = None,
    url: Optional[str] = None,
    emoji: Optional[Emoji] = None,
    disabled: bool = False,
    **kwargs,
) -> B:
    """
    A helper function that passes arguments to `Button`.

    Previous:

    ```py
    button = Button(style=1, label="1", custom_id="1", ...)
    ```

    Now:

    ```py
    button = Button(1, "1", custom_id="1", ...)
    ```

    Parameters:

    * `style: ButtonStyle | int`: The style of the button.
    * `label: str`: The label of the button.
    * `(?)custom_id: str`: The custom id of the button. *Required if the button is not a link.*
    * `(?)url: str`: The URL of the button. *Required if the button is a link.*
    * `?emoji: Emoji`: The emoji of the button.
    * `?disabled: bool = False`: Whether the button is disabled.
    * `**kwargs: dict`: Any additional arguments of the button.

    Returns:

    `Button`
    """
    log.debug(
        f"Creating Button with {style=}, {label=}, {custom_id=}, {url=}, {emoji=}, {disabled=}"
    )
    if custom_id and url:
        raise ValueError("`custom_id` and `url` cannot be used together!")

    if not (custom_id or url):
        raise ValueError("`custom_id` or `url` must be specified!")

    if style == ButtonStyle.LINK and not url:
        raise ValueError("`url` must be specified if `style` is `ButtonStyle.LINK`!")
    if url and style != ButtonStyle.LINK:
        raise ValueError("`url` can only be specified if `style` is `ButtonStyle.LINK`!")

    if style != ButtonStyle.LINK and not custom_id:
        raise ValueError("`custom_id` must be specified if `style` is not `ButtonStyle.LINK`!")
    if custom_id and style == ButtonStyle.LINK:
        raise ValueError("`custom_id` can only be specified if `style` is not `ButtonStyle.LINK`!")

    return B(
        style=style,
        label=label,
        custom_id=custom_id,
        url=url,
        emoji=emoji,
        disabled=disabled,
        **kwargs,
    )


def SelectOption(
    label: str,
    value: str,
    description: Optional[str] = None,
    emoji: Optional[Emoji] = None,
    disabled: bool = False,
    **kwargs,
) -> SO:
    """
    A helper function that passes arguments to `SelectOption`.

    Before:

    ```py
    option = SelectOption(label="1", value="1", ...)
    ```

    Now:

    ```py
    option = SelectOption("1", "1", ...)
    ```

    Parameters:

    * `label: str`: The label of the option.
    * `value: str`: The value of the option.
    * `?description: str`: The description of the option.
    * `?emoji: Emoji`: The emoji of the option.
    * `?disabled: bool = False`: Whether the option is disabled.
    * `**kwargs: dict`: Any additional arguments of the option.

    Returns:

    `SelectOption`
    """
    log.debug(
        f"Creating SelectOption with {label=}, {value=}, {description=}, {emoji=}, {disabled=}"
    )
    return SO(
        label=label,
        value=value,
        description=description,
        emoji=emoji,
        disabled=disabled,
        **kwargs,
    )


def SelectMenu(
    custom_id: str,
    options: List[SO],
    *,
    placeholder: Optional[str] = None,
    min_values: Optional[int] = None,
    max_values: Optional[int] = None,
    disabled: bool = False,
    **kwargs,
) -> SM:
    """
    A helper function that passes arguments to `SelectMenu`.

    Previous:

    ```py
    select = SelectMenu(custom_id="s", options=[...], ...)
    ```

    Now:

    ```py
    select = SelectMenu("s", [...], ...)
    ```

    Parameters:

    * `custom_id: str`: The custom id of the select menu.
    * `options: list[SelectOption]`: The options of the select menu.
    * `?placeholder: str`: The placeholder of the select menu.
    * `?min_values: int`: The minimum number of values that can be selected.
    * `?max_values: int`: The maximum number of values that can be selected.
    * `?disabled: bool`: Whether the select menu is disabled. Defaults to `False`.
    * `**kwargs: dict`: Any additional arguments of the select menu.

    Returns:

    `SelectMenu`
    """
    log.debug(
        f"Creating SelectMenu with {custom_id=}, {options=}, {placeholder=}, {min_values=}, {max_values=}, {disabled=}"
    )
    return SM(
        custom_id=custom_id,
        options=options,
        placeholder=placeholder,
        min_values=min_values,
        max_values=max_values,
        disabled=disabled,
        **kwargs,
    )


def TextInput(
    custom_id: str,
    label: str,
    style: TST = TST.SHORT,
    value: Optional[str] = None,
    required: bool = True,
    placeholder: Optional[str] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    **kwargs,
) -> TI:
    """
    A helper function that passes arguments to `TextInput`.

    Before:

    ```py
    ti = TextInput(custom_id="ti", label="ti", style=1, ...)
    ```

    Now:

    ```py
    ti = TextInput("ti", "ti", 1, ...)
    ```

    Parameters:

    * `custom_id: str`: The custom id of the text input.
    * `label: str`: The label of the text input.
    * `?style: TextInputStyle | int`: The style of the text input.
    * `?value: str`: The value of the text input.
    * `?required: bool = True`: Whether the text input is required.
    * `?placeholder: str`: The placeholder of the text input.
    * `?min_length: int`: The minimum length of the text input.
    * `?max_length: int`: The maximum length of the text input.

    Returns:

    `TextInput`
    """
    log.debug(
        f"Creating TextInput with {custom_id=}, {label=}, {style=}, {value=}, {required=}, {placeholder=}, {min_length=}, {max_length=}"
    )
    return TI(
        custom_id=custom_id,
        label=label,
        style=style,
        value=value,
        required=required,
        placeholder=placeholder,
        min_length=min_length,
        max_length=max_length,
        **kwargs,
    )


def Modal(custom_id: str, title: str, components: List[TI], **kwargs) -> M:
    """
    A helper function that passes arguments to `Modal`.

    Before:

    ```py
    modal = Modal(custom_id="modal", title="Modal", components=[...])
    ```

    Now:

    ```py
    modal = Modal("modal", "Modal", [...])
    ```

    Parameters:

    * `custom_id: str`: The custom id of the modal.
    * `title: str`: The title of the modal.
    * `components: list[TextInput]`: The components of the modal.
    * `**kwargs: dict`: Any additional arguments of the modal.

    Returns:

    `Modal`
    """
    log.debug(f"Creating Modal with {custom_id=}, {title=}, {components=}")
    return M(custom_id=custom_id, title=title, components=components, **kwargs)



"""
extension

Content:

* base: base class for extension.
* version: version of the extension.

GitHub: https://github.com/interactions-py/enhanced/blob/main/interactions/ext/enhanced/extension.py

(c) 2022 interactions-py.
"""
import types
from logging import Logger
from re import fullmatch
from typing import Union

from interactions import Client, CommandContext, ComponentContext, Extension
from interactions.ext import Base, Version, VersionAuthor

from ._logging import get_logger

__all__ = ("Enhanced", "setup")

log: Logger = get_logger("extension")


version: Version = Version(
    version="4.0.0",
    author=VersionAuthor(
        name="Toricane",
        email="prjwl028@gmail.com",
    ),
)

base = Base(
    name="enhanced",
    version=version,
    description="Enhanced interactions for interactions.py",
    link="https://github.com/interactions-py/enhanced",
    packages=["interactions.ext.enhanced"],
    requirements=[
        "discord-py-interactions>=4.3.0",
        "typing_extensions",
    ],
)


class Enhanced(Extension):
    """
    This is the core of this library, initialized when loading the extension.

    It applies hooks to the client for additional and modified features.

    ```py
    # main.py
    client.load("interactions.ext.enhanced", ...)  # optional args/kwargs
    ```

    Parameters:

    * `(?)bot: Client`: The client instance. Not required if using `client.load("interactions.ext.enhanced", ...)`.
    * `?ignore_warning: bool`: Whether to ignore the warning. Defaults to `False`.
    * `?modify_callbacks: bool`: Whether to modify callback decorators. Defaults to `True`.
    """

    def __init__(
        self,
        bot: Client,
        *,
        ignore_warning: bool = False,
        modify_callbacks: bool = True,
    ):
        if not isinstance(bot, Client):
            log.critical("The bot is not an instance of Client")
            if not ignore_warning:
                raise TypeError(f"{bot.__class__.__name__} is not interactions.Client!")
        log.debug("The bot is an instance of Client")

        if modify_callbacks:
            from .callbacks import component, modal

            log.debug("Modifying component callbacks (modify_callbacks)")
            bot.component = types.MethodType(component, bot)

            bot.event(self._on_component, name="on_component")
            log.debug("Registered on_component")

            log.debug("Modifying modal callbacks (modify_callbacks)")
            bot.modal = types.MethodType(modal, bot)

            bot.event(self._on_modal, name="on_modal")
            log.debug("Registered on_modal")

        log.info("Hooks applied")

    async def __callback(self, ctx: Union[ComponentContext, CommandContext]):
        callback = "component" if isinstance(ctx, ComponentContext) else "modal"
        websocket = self.client._websocket

        if not any(
            any(hasattr(func, "startswith") or hasattr(func, "regex") for func in funcs)
            for _, funcs in websocket._dispatch.events.items()
        ):
            return

        for decorator_custom_id, funcs in websocket._dispatch.events.items():
            for func in funcs:
                if hasattr(func, "startswith"):
                    if ctx.data.custom_id.startswith(
                        decorator_custom_id.replace(f"{callback}_startswith_", "")
                    ):
                        log.info(f"{func} startswith {func.startswith} matched")
                        return websocket._dispatch.dispatch(decorator_custom_id, ctx)
                elif hasattr(func, "regex") and fullmatch(
                    func.regex,
                    ctx.data.custom_id.replace(f"{callback}_regex_", ""),
                ):
                    log.info(f"{func} regex {func.regex} matched")
                    return websocket._dispatch.dispatch(decorator_custom_id, ctx)

    async def _on_component(self, ctx: ComponentContext):
        """on_component callback for modified callbacks."""
        return await self.__callback(ctx)

    async def _on_modal(self, ctx: CommandContext):
        """on_modal callback for modified callbacks."""
        return await self.__callback(ctx)


def setup(
    bot: Client,
    *,
    ignore_warning: bool = False,
    modify_callbacks: bool = True,
) -> Enhanced:
    """
    This function initializes the core of the library, `Enhanced`.

    It applies hooks to the client for additional and modified features.

    ```py
    # main.py
    client.load("interactions.ext.enhanced", ...)  # optional args/kwargs
    ```

    Parameters:

    * `(?)client: Client`: The client instance. Not required if using `client.load("interactions.ext.enhanced", ...)`.
    * `?ignore_warning: bool`: Whether to ignore the warning. Defaults to `False`.
    * `?modify_callbacks: bool`: Whether to modify callback decorators. Defaults to `True`.
    """
    log.info("Setting up Enhanced")
    return Enhanced(
        bot,
        ignore_warning=ignore_warning,
        modify_callbacks=modify_callbacks,
    )



"""
commands

Content:

* setup_options: Sets up the options of the command.

GitHub: https://github.com/interactions-py/enhanced/blob/main/interactions/ext/enhanced/commands.py

(c) 2022 interactions-py.
"""
from inspect import signature
from logging import Logger
from typing import Awaitable, Callable

from typing_extensions import _AnnotatedAlias

from ._logging import get_logger
from .command_models import EnhancedOption, parameters_to_options

__all__ = ("setup_options",)

log: Logger = get_logger("command")


def setup_options(coro: Callable[..., Awaitable]):
    """
    Sets up the options of the command.

    Usage:
    ```py
    @bot.command()
    @setup_options
    async def test(ctx, option: EnhancedOption(...)):
        ...
    ```

    Parameters:

    * `(X)coro: Callable[..., Awaitable]`: The coroutine to setup the options of.
    """
    params = signature(coro).parameters

    def check(num: int):
        nonlocal params
        return len(params) > num and all(
            isinstance(param.annotation, (EnhancedOption, _AnnotatedAlias))
            for param in list(params.values())[num:]
        )

    if not (check(1) or check(2)):
        return coro

    options = parameters_to_options(coro)

    if hasattr(coro, "_options") and isinstance(coro._options, list):
        coro._options.extend(options)
    else:
        coro._options = options

    return coro



"""
command_models

Content:

* EnhancedOption: typehintable option
* option: decoratable option

GitHub: https://github.com/interactions-py/enhanced/blob/main/interactions/ext/enhanced/command_models.py

(c) 2022 interactions-py.
"""
from inspect import _empty, signature
from typing import TYPE_CHECKING, Awaitable, Callable, List, Optional, Type, Union, get_args

from interactions import MISSING, Attachment, Channel, File, Member, Option, OptionType, Role, User

from ._logging import get_logger

if TYPE_CHECKING:
    from types import MappingProxyType
    from inspect import Parameter

from typing_extensions import _AnnotatedAlias

__all__ = ("EnhancedOption",)

log = get_logger("command_models")
_type: type = type


def get_type(param):
    """Gets the type of the parameter."""
    return get_args(param.annotation)[0] or get_args(param.annotation)[1].type


def get_option(param):
    """Gets the `EnhancedOption` of the parameter."""
    return get_args(param.annotation)[1]


def type_to_int(param):
    """Converts the type to an integer."""
    type: Union[_type, int, OptionType] = get_type(param)
    if isinstance(type, int):
        return type
    if type in (str, int, float, bool):
        if type is str:
            return OptionType.STRING
        if type is int:
            return OptionType.INTEGER
        if type is float:
            return OptionType.NUMBER
        if type is bool:
            return OptionType.BOOLEAN
    elif isinstance(type, OptionType):
        return type
    elif type is User or type is Member:
        return OptionType.USER
    elif type is Channel:
        return OptionType.CHANNEL
    elif type is Role:
        return OptionType.ROLE
    else:
        raise TypeError(f"Invalid type: {type}")


class EnhancedOption:
    """
    An alternative way of providing options by typehinting.

    Basic example:

    ```py
    @bot.command(...)
    async def command(ctx, name: EnhancedOption(int, "description") = 5):
        ...
    ```

    Full-blown example:

    ```py
    from interactions import OptionType, Channel
    from interactions.ext.enhanced import EnhancedOption
    from typing_extensions import Annotated

    @bot.command()
    async def options(
        ctx,
        option1: Annotated[str, EnhancedOption(description="...")],
        option2: Annotated[OptionType.MENTIONABLE, EnhancedOption(description="...")],
        option3: Annotated[Channel, EnhancedOption(description="...")],
    ):
        \"""Says something!\"""
        await ctx.send("something")
    ```

    Parameters:

    * `?option_type: type | int | OptionType`: The type of the option.
    * `?description: str = "No description"`: The description of the option.
    * `?name: str`: The name of the option. Defaults to the argument name.
    * `?**kwargs`: Any additional keyword arguments, same as `ipy.Option`.
    """

    def __init__(
        self,
        option_type: Union[_type, int, OptionType] = None,
        /,
        description: str = "No description",
        name: Optional[str] = None,
        **kwargs,
    ):
        log.debug("EnhancedOption.__init__")
        if isinstance(option_type, (int, _type(None))):
            self.type = option_type
        elif option_type in (str, int, float, bool):
            if option_type is str:
                self.type = OptionType.STRING
            elif option_type is int:
                self.type = OptionType.INTEGER
            elif option_type is float:
                self.type = OptionType.NUMBER
            elif option_type is bool:
                self.type = OptionType.BOOLEAN
        elif isinstance(option_type, OptionType):
            self.type = option_type
        elif option_type is User or option_type is Member:
            self.type = OptionType.USER
        elif option_type is Channel:
            self.type = OptionType.CHANNEL
        elif option_type is Role:
            self.type = OptionType.ROLE
        elif option_type is File or option_type is Attachment:
            self.type = OptionType.ATTACHMENT
        else:
            raise TypeError(f"Invalid type: {option_type}")

        self.description = description or "No description"
        self.name = name
        self.kwargs = kwargs

    def __repr__(self):
        return f"<EnhancedOption type={self.type}, name={self.name}>"


def loop_params(params: dict, stop: int) -> dict:
    """Loops through the parameters and deletes until stop index."""
    for i, key in enumerate(params.copy()):
        if i > stop:
            break
        del params[key]
    return params


def format_parameters(coro: Callable[..., Awaitable]):
    """Formats the parameters of a function."""
    params: MappingProxyType[str, Parameter] = signature(coro).parameters
    _params: dict = dict(params.items())

    if "." in coro.__qualname__:
        return loop_params(_params, 1)
    else:
        return loop_params(_params, 0)


def parameters_to_options(
    coro: Callable[..., Awaitable], has_res: bool = False
) -> Optional[List[Option]]:
    """Converts `EnhancedOption`s to `Option`s."""
    log.debug("parameters_to_options:")
    params: dict = format_parameters(coro)
    if has_res:
        for key in params:
            del params[key]
            break

    _options: List[Union[Option, Type[MISSING]]] = [
        Option(
            type=param.annotation.type,
            name=param.annotation.name or __name,
            description=param.annotation.description,
            required=param.default is _empty,
            **param.annotation.kwargs,
        )
        if isinstance(param.annotation, EnhancedOption)
        else Option(
            type=type_to_int(param),
            name=get_option(param).name or __name,
            description=get_option(param).description,
            required=param.default is _empty,
            **get_option(param).kwargs,
        )
        if isinstance(param.annotation, _AnnotatedAlias)
        else MISSING
        for __name, param in params.items()
    ]

    if any(opt is MISSING for opt in _options):
        raise TypeError(
            "You must typehint with `EnhancedOption` or specify `options=...` in the decorator!"
        )
    log.debug(f"  _options: {_options}\n")

    return _options



"""
callbacks

Content:

* component: component callback
* modal: modal callback
* extension_component: extension component callback
* extension_modal: extension modal callback

GitHub: https://github.com/interactions-py/enhanced/blob/main/interactions/ext/enhanced/callbacks.py

(c) 2022 interactions-py.
"""
from re import compile
from typing import Awaitable, Callable, Optional, Union

from interactions import Button, Client, Component, Modal, SelectMenu

from ._logging import get_logger

log = get_logger("callback")
Coroutine = Callable[..., Awaitable]

__all__ = ("component", "modal", "extension_component", "extension_modal")


def component(
    bot: Client,
    component: Union[str, Button, SelectMenu],
    startswith: bool = False,
    regex: bool = False,
) -> Callable[[Coroutine], Coroutine]:
    """
    A modified decorator that allows you to add more information to the `custom_id` and use
    `startswith` or `regex` to invoke the callback.

    ```py
    bot.load("interactions.ext.enhanced")

    # startswith:
    @bot.component("test", startswith=True)
    async def test(ctx):
        ...

    # regex:
    @bot.component(r"^[a-z0-9_-]{1,32}$", regex=True)
    async def test(ctx):
        ...
    ```

    The startswith callback is called if the `custom_id` starts with the given string.

    The regex callback is called if the `custom_id` matches the given regex.

    Parameters:

    * `(X)bot: Client`: The bot client.
    * `component: str | Button | SelectMenu`: The component custom_id or regex to listen to.
    * `?startswith: bool = False`: Whether the component custom_id starts with the given string.
    * `?regex: bool = False`: Whether the component custom_id matches the given regex.
    """

    def decorator(coro: Coroutine) -> Coroutine:
        if hasattr(coro, "__extension"):
            return bot.event(coro, name=f"component_{component}")

        payload: str = (
            Component(**component._json).custom_id
            if isinstance(component, (Button, SelectMenu))
            else component
        )
        if startswith and regex:
            log.error("Cannot use both startswith and regex.")
            raise ValueError("Cannot use both startswith and regex!")

        if startswith:
            try:
                coro.startswith = True
            except AttributeError:
                coro.__func__.startswith = True
            bot.event(coro, name=f"component_startswith_{payload}")
        elif regex:
            try:
                coro.regex = compile(payload)
            except AttributeError:
                coro.__func__.regex = compile(payload)
            bot.event(coro, name=f"component_regex_{payload}")
        else:
            bot.event(coro, name=f"component_{payload}")

        log.debug(f"Component callback, {startswith=}, {regex=}")
        return coro

    return decorator


def modal(
    bot: Client,
    modal: Union[Modal, str],
    startswith: bool = False,
    regex: bool = False,
) -> Callable[[Coroutine], Coroutine]:
    """
    A modified decorator that allows you to add more information to the `custom_id` and use
    `startswith` or `regex` to invoke the callback.

    ```py
    bot.load("interactions.ext.enhanced")

    # startswith:
    @bot.modal("test", startswith=True)
    async def test(ctx):
        ...

    # regex:
    @bot.modal(r"^[a-z0-9_-]{1,32}$", regex=True)
    async def test(ctx):
        ...
    ```

    The startswith callback is called if the `custom_id` starts with the given string.

    The regex callback is called if the `custom_id` matches the given regex.

    Parameters:

    * `(X)bot: Client`: The bot client.
    * `modal: str | Modal`: The modal custom_id or regex to listen to.
    * `?startswith: bool = False`: Whether the modal custom_id starts with the given string.
    * `?regex: bool`: Whether the modal custom_id matches the given regex.
    """

    def decorator(coro: Coroutine) -> Coroutine:
        if hasattr(coro, "__extension"):
            return bot.event(coro, name=f"modal_{modal}")

        payload: str = modal.custom_id if isinstance(modal, Modal) else modal
        if startswith and regex:
            log.error("Cannot use both startswith and regex.")
            raise ValueError("Cannot use both startswith and regex!")

        if startswith:
            try:
                coro.startswith = True
            except AttributeError:
                coro.__func__.startswith = True
            bot.event(coro, name=f"modal_startswith_{payload}")
        elif regex:
            try:
                coro.regex = compile(payload)
            except AttributeError:
                coro.__func__.regex = compile(payload)
            bot.event(coro, name=f"modal_regex_{payload}")
        else:
            bot.event(coro, name=f"modal_{payload}")

        log.debug(f"Modal callback, {startswith=}, {regex=}")
        return coro

    return decorator


def extension_component(
    component: Union[str, Button, SelectMenu],
    startswith: Optional[bool] = False,
    regex: Optional[bool] = False,
):
    """
    A modified decorator that allows you to add more information to the `custom_id` and use
    `startswith` or `regex` to invoke the callback inside of `Extension`s.

    ```py
    # main.py:
    bot.load("interactions.ext.enhanced")

    # startswith:
    @extension_component("test", startswith=True)
    async def test(self, ctx):
        ...

    # regex:
    @extension_component(r"^[a-z0-9_-]{1,32}$", regex=True)
    async def test(self, ctx):
        ...
    ```

    The startswith callback is called if the `custom_id` starts with the given string.

    The regex callback is called if the `custom_id` matches the given regex.

    Parameters:

    * `component: str | Button | SelectMenu`: The component custom_id or regex to listen to.
    * `?startswith: bool = False`: Whether the component custom_id starts with the given string.
    * `?regex: bool = False`: Whether the component custom_id matches the given regex.
    """

    def decorator(func):
        if startswith and regex:
            log.error("Cannot use both startswith and regex.")
            raise ValueError("Cannot use both startswith and regex!")

        func.__extension = True
        payload: str = (
            Component(**component._json).custom_id
            if isinstance(component, (Button, SelectMenu))
            else component
        )

        if startswith:
            try:
                func.startswith = True
            except AttributeError:
                func.__func__.startswith = True
            payload = f"startswith_{payload}"
        elif regex:
            try:
                func.regex = compile(payload)
            except AttributeError:
                func.__func__.regex = compile(payload)
            payload = f"regex_{payload}"

        log.debug(f"Extension component callback, {startswith=}, {regex=}")

        func.__component_data__ = (
            (),
            {"component": payload, "startswith": startswith, "regex": regex},
        )
        return func

    return decorator


def extension_modal(
    modal: Union[Modal, str],
    startswith: bool = False,
    regex: bool = False,
):
    """
    A modified decorator that allows you to add more information to the `custom_id` and use
    `startswith` or `regex` to invoke the callback inside of `Extension`s.

    ```py
    # main.py:
    bot.load("interactions.ext.enhanced")

    # startswith:
    @extension_modal("test", startswith=True)
    async def test(self, ctx):
        ...

    # regex:
    @extension_modal(r"^[a-z0-9_-]{1,32}$", regex=True)
    async def test(self, ctx):
        ...
    ```

    The startswith callback is called if the `custom_id` starts with the given string.

    The regex callback is called if the `custom_id` matches the given regex.

    Parameters:

    * `modal: str | Modal`: The modal custom_id or regex to listen to.
    * `?startswith: bool = False`: Whether the modal custom_id starts with the given string.
    * `?regex: bool` = False: Whether the modal custom_id matches the given regex.
    """

    def decorator(func):
        if startswith and regex:
            log.error("Cannot use both startswith and regex.")
            raise ValueError("Cannot use both startswith and regex!")

        func.__extension = True
        payload: str = modal.custom_id if isinstance(modal, Modal) else modal

        if startswith:
            try:
                func.startswith = True
            except AttributeError:
                func.__func__.startswith = True
            payload = f"startswith_{payload}"
        elif regex:
            try:
                func.regex = compile(payload)
            except AttributeError:
                func.__func__.regex = compile(payload)
            payload = f"regex_{payload}"

        log.debug(f"Extension modal callback, {startswith=}, {regex=}")

        func.__modal_data__ = (
            (),
            {"modal": payload, "startswith": startswith, "regex": regex},
        )
        return func

    return decorator



"""
_logging

Content:

* Data: constants
* CustomFormatter: custom formatter for logging
* get_logger: gets a logger for the library

GitHub: https://github.com/interactions-py/enhanced/blob/main/interactions/ext/enhanced/_logging.py

(c) 2022 interactions-py.
"""
import logging
from typing import ClassVar, List, Optional, Union

from colorama import Fore, Style, init


class Data:
    """A class representing constants for the library."""

    LOG_LEVEL: ClassVar[int] = logging.ERROR
    LOGGERS: List[str] = []


class CustomFormatter(logging.Formatter):
    """A class that allows for customized logged outputs from the library."""

    format_str: str = "%(levelname)s:%(name)s:(ln.%(lineno)d):%(message)s"
    formats: dict = {
        logging.DEBUG: Fore.CYAN + format_str + Fore.RESET,
        logging.INFO: Fore.GREEN + format_str + Fore.RESET,
        logging.WARNING: Fore.YELLOW + format_str + Fore.RESET,
        logging.ERROR: Fore.RED + format_str + Fore.RESET,
        logging.CRITICAL: Style.BRIGHT + Fore.RED + format_str + Fore.RESET + Style.NORMAL,
    }

    def __init__(self):
        super().__init__()
        init(autoreset=True)

    def format(self, record):
        log_format = self.formats.get(record.levelno)
        formatter = logging.Formatter(log_format)
        return formatter.format(record)


def get_logger(
    logger: Optional[Union[logging.Logger, str]] = None,
    handler: Optional[logging.Handler] = logging.StreamHandler(),
) -> logging.Logger:
    """Gets a logger for the library."""
    _logger = logging.getLogger(logger) if isinstance(logger, str) else logger
    _logger_name = logger if isinstance(logger, str) else logger.name
    if len(_logger.handlers) > 1:
        _logger.removeHandler(_logger.handlers[0])
    _handler = handler
    _handler.setFormatter(CustomFormatter())
    _handler.setLevel(Data.LOG_LEVEL)
    _logger.addHandler(_handler)
    _logger.propagate = True

    Data.LOGGERS.append(_logger_name)
    return _logger



"""
enhanced

Enhanced interactions for interactions.py.

Everything within and including the below modules
is importable directly from enhanced.

Modules:

* callbacks: component or modal callbacks.
* commands: slash commands.
* command_models: slash command option models.
* components: components.
* cooldowns: command cooldowns.
* extension: extension.
* subcommands: subcommands.

GitHub: https://github.com/interactions-py/enhanced/

PyPI: https://pypi.org/project/enhanced/

(c) 2022 interactions-py.
"""
from . import _logging, callbacks, command_models, commands, components, cooldowns, extension
from ._logging import CustomFormatter, Data, get_logger
from .callbacks import component, extension_component, extension_modal, modal
from .command_models import EnhancedOption
from .commands import setup_options
from .components import ActionRow, Button, Modal, SelectMenu, TextInput
from .cooldowns import cooldown
from .extension import Enhanced, base, setup, version

# fmt: off
__all__ = [
    "_logging",
        "Data",  # noqa E131
        "CustomFormatter",
        "get_logger",
    # "cmd",
        "command_models",
            "EnhancedOption",  # noqa E131
        "commands",
            "setup_options",
    # "cmpt",
        "callbacks",
            "component",
            "modal",
            "extension_component",
            "extension_modal",
        "components",
            "ActionRow",
            "Button",
            "SelectMenu",
            "TextInput",
            "Modal",
    "extension",
        "Enhanced",
        "setup",
        "base",
        "version",
    "cooldowns",
        "cooldown",
]
# fmt: on
